# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========

import base64
import hashlib
import hmac
import os
import re
import time
import urllib.parse
from typing import Any, Dict, List, Literal, Optional

import requests

from camel.logger import get_logger
from camel.toolkits import FunctionTool
from camel.toolkits.base import BaseToolkit
from camel.utils import MCPServer, api_keys_required, retry_on_error

logger = get_logger(__name__)

# Global variables for caching access token
_dingtalk_access_token = None
_dingtalk_access_token_expires_at = 0


@retry_on_error()
def _get_dingtalk_access_token() -> str:
    r"""Gets access token for Dingtalk API.

    Returns:
        str: Access token for API requests.
    """
    global _dingtalk_access_token, _dingtalk_access_token_expires_at

    if (
        _dingtalk_access_token
        and _dingtalk_access_token_expires_at > time.time()
    ):
        return _dingtalk_access_token

    app_key = os.environ.get("DINGTALK_APP_KEY", "")
    app_secret = os.environ.get("DINGTALK_APP_SECRET", "")

    url = "https://oapi.dingtalk.com/gettoken"
    params = {"appkey": app_key, "appsecret": app_secret}

    response = requests.get(url, params=params, timeout=30)
    response.raise_for_status()
    data = response.json()

    if data.get("errcode") == 0 and "access_token" in data:
        _dingtalk_access_token = data["access_token"]
        _dingtalk_access_token_expires_at = (
            time.time() + data.get("expires_in", 7200) - 60
        )
        logger.info("Dingtalk access token refreshed.")
        return _dingtalk_access_token
    else:
        errcode = data.get("errcode")
        errmsg = data.get("errmsg", "Unknown error")
        raise ValueError(f"Failed to get access token {errcode}: {errmsg}")


def _make_dingtalk_request(
    method: Literal["GET", "POST"], endpoint: str, **kwargs
) -> Dict[str, Any]:
    r"""Makes authenticated request to Dingtalk API.

    Args:
        method (Literal["GET", "POST"]): HTTP method to use.
        endpoint (str): API endpoint path.
        **kwargs: Additional arguments passed to requests.

    Returns:
        Dict[str, Any]: API response data.

    Raises:
        Exception: If API request fails or returns error.
    """
    global _dingtalk_access_token, _dingtalk_access_token_expires_at
    access_token = _get_dingtalk_access_token()

    # Handle URL parameter concatenation
    separator = "&" if "?" in endpoint else "?"
    url = f"https://oapi.dingtalk.com{endpoint}{separator}access_token={access_token}"

    if method.upper() == "GET":
        response = requests.get(url, **kwargs)
    else:
        response = requests.post(url, **kwargs)

    response.raise_for_status()
    data = response.json()

    if data.get("errcode") and data.get("errcode") != 0:
        errcode = data.get("errcode")
        errmsg = data.get("errmsg", "Unknown error")
        raise ValueError(f"Dingtalk API error {errcode}: {errmsg}")

    return data


def _generate_signature(secret: str, timestamp: str) -> str:
    r"""Generates signature for Dingtalk webhook.

    Args:
        secret (str): Webhook secret.
        timestamp (str): Current timestamp.

    Returns:
        str: Generated signature.
    """
    string_to_sign = f"{timestamp}\n{secret}"
    hmac_code = hmac.new(
        secret.encode("utf-8"), string_to_sign.encode("utf-8"), hashlib.sha256
    ).digest()
    return base64.b64encode(hmac_code).decode('utf-8')


@MCPServer()
class DingtalkToolkit(BaseToolkit):
    r"""A toolkit for Dingtalk operations.

    This toolkit provides methods to interact with the Dingtalk API,
    allowing users to send messages, manage users, departments, and handle
    webhook operations.
    """

    def __init__(self, timeout: Optional[float] = None):
        r"""Initializes the DingtalkToolkit.

        Args:
            timeout (Optional[float]): Timeout for API requests in seconds.
        """
        super().__init__(timeout=timeout)
        self.base_url = "https://oapi.dingtalk.com"

        # Validate credentials
        app_key = os.environ.get("DINGTALK_APP_KEY", "")
        app_secret = os.environ.get("DINGTALK_APP_SECRET", "")

        if not all([app_key, app_secret]):
            raise ValueError(
                "Dingtalk credentials missing. Set DINGTALK_APP_KEY and"
                " DINGTALK_APP_SECRET."
            )

        # Initialize access token for enterprise reliability
        self._initialize_token_safely()

    def _initialize_token_safely(self):
        r"""Safely initializes access token during toolkit setup.

        This method attempts to get an access token during initialization
        but doesn't raise exceptions if it fails, allowing the toolkit
        to be instantiated even if credentials are temporarily invalid.
        """
        try:
            _get_dingtalk_access_token()
            logger.info("Dingtalk toolkit initialized successfully.")
        except Exception as e:
            logger.warning(f"Failed to initialize access token: {e}")
            logger.warning(
                "Toolkit created but may fail on actual API calls. "
                "Check your credentials."
            )

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_text_message(
        self,
        userid: str,
        content: str,
    ) -> str:
        r"""Sends a text message to a Dingtalk user.

        Args:
            userid (str): The user's userid.
            content (str): Message content.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-single-chat-message
        """
        payload = {
            "msg": {"msgtype": "text", "text": {"content": content}},
            "userid": userid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"Message sent successfully to user {userid}."
        except Exception as e:
            return f"Failed to send message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_markdown_message(
        self,
        userid: str,
        title: str,
        markdown_content: str,
    ) -> str:
        r"""Sends a markdown message to a Dingtalk user.

        Args:
            userid (str): The user's userid.
            title (str): Message title.
            markdown_content (str): Markdown formatted content.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-single-chat-message
        """
        payload = {
            "msg": {
                "msgtype": "markdown",
                "markdown": {"title": title, "text": markdown_content},
            },
            "userid": userid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"Markdown message sent successfully to user {userid}."
        except Exception as e:
            return f"Failed to send markdown message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_user_info(
        self,
        userid: str,
    ) -> Dict[str, Any]:
        r"""Retrieves Dingtalk user information.

        Args:
            userid (str): The user's userid.

        Returns:
            Dict[str, Any]: User information or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-user-details
        """
        payload = {"userid": userid}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/v2/user/get",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get user info: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_department_list(
        self,
        dept_id: Optional[int] = None,
    ) -> Dict[str, Any]:
        r"""Retrieves list of departments.

        Args:
            dept_id (Optional[int]): Department ID. If None, gets root
                departments.

        Returns:
            Dict[str, Any]: Department list or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/obtain-the-department-list-v2
        """
        payload = {}
        if dept_id is not None:
            payload["dept_id"] = dept_id

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/v2/department/listsub",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get department list: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_department_users(
        self,
        dept_id: int,
        offset: int = 0,
        size: int = 100,
    ) -> Dict[str, Any]:
        r"""Retrieves users in a department.

        Args:
            dept_id (int): Department ID.
            offset (int): Offset for pagination (default: 0).
            size (int): Number of users to retrieve (default: 100, max: 100).

        Returns:
            Dict[str, Any]: Users list or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/queries-the-complete-information-of-a-department-user
        """
        payload = {"dept_id": dept_id, "offset": offset, "size": size}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/v2/user/list",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get department users: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_search_users_by_name(
        self,
        name: str,
    ) -> Dict[str, Any]:
        r"""Searches for users by name.

        Args:
            name (str): User name to search for.

        Returns:
            Dict[str, Any]: Search results or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-users
        """
        payload = {"name": name}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/user/search",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to search users: {e!s}"}

    # Include more webhook methods if useful. Double check link implementation
    # is done correctly here.
    def dingtalk_send_webhook_message(
        self,
        content: str,
        msgtype: Literal["text", "markdown", "link", "actionCard"] = "text",
        title: Optional[str] = None,
        webhook_url: Optional[str] = None,
        webhook_secret: Optional[str] = None,
    ) -> str:
        r"""Sends a message via Dingtalk webhook.

        Args:
            content (str): Message content.
            msgtype (Literal): Message type (text, markdown, link, actionCard).
            title (Optional[str]): Message title (required for markdown).
            webhook_url (Optional[str]): Webhook URL. If None, uses env var.
            webhook_secret (Optional[str]): Webhook secret. If None, uses env
                var.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/robots/custom-robot-access
        """
        # Get webhook configuration
        url = webhook_url or os.environ.get("DINGTALK_WEBHOOK_URL", "")
        secret = webhook_secret or os.environ.get(
            "DINGTALK_WEBHOOK_SECRET", ""
        )

        if not url:
            return "Error: Webhook URL not provided or set in environment"

        # Prepare message payload
        payload: Dict[str, Any] = {"msgtype": msgtype}

        if msgtype == "text":
            payload["text"] = {"content": content}
        elif msgtype == "markdown":
            if not title:
                return "Error: Title is required for markdown messages"
            payload["markdown"] = {"title": title, "text": content}
        elif msgtype == "link":
            # For link messages, content should be structured differently
            # This is a simplified implementation
            payload["link"] = {
                "text": content,
                "title": title or "Link Message",
                "messageUrl": "https://www.dingtalk.com/",
            }
        elif msgtype == "actionCard":
            payload["actionCard"] = {
                "title": title or "Action Card",
                "text": content,
                "singleTitle": "Read More",
                "singleURL": "https://www.dingtalk.com/",
            }

        # Add signature if secret is provided
        if secret:
            timestamp = str(round(time.time() * 1000))
            sign = _generate_signature(secret, timestamp)
            url += f"&timestamp={timestamp}&sign={urllib.parse.quote(sign)}"

        try:
            response = requests.post(
                url,
                json=payload,
                headers={"Content-Type": "application/json"},
                timeout=self.timeout,
            )
            response.raise_for_status()

            result = response.json()
            if result.get("errcode") == 0:
                return "Webhook message sent successfully"
            else:
                return (
                    f"Webhook error: {result.get('errmsg', 'Unknown error')}"
                )

        except Exception as e:
            return f"Failed to send webhook message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_create_group(
        self,
        name: str,
        owner: str,
        useridlist: List[str],
    ) -> Dict[str, Any]:
        r"""Creates a Dingtalk group.

        Args:
            name (str): Group name.
            owner (str): Group owner's userid.
            useridlist (List[str]): List of user IDs to add to the group.

        Returns:
            Dict[str, Any]: Group creation result with chatid or error.

        References:
            https://open.dingtalk.com/document/orgapp-server/create-group-session
        """
        payload = {
            "name": name,
            "owner": owner,
            "useridlist": useridlist,
        }

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/im/chat/create",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to create group: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_group_message(
        self,
        chatid: str,
        content: str,
        msgtype: Literal["text", "markdown"] = "text",
    ) -> str:
        r"""Sends a message to a Dingtalk group.

        Args:
            chatid (str): Group chat ID.
            content (str): Message content.
            msgtype (Literal["text", "markdown"]): Message type.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-group-messages
        """
        msg_data: Dict[str, Any] = {"msgtype": msgtype}

        if msgtype == "text":
            msg_data["text"] = {"content": content}
        elif msgtype == "markdown":
            msg_data["markdown"] = {"text": content}

        payload = {
            "msg": msg_data,
            "chatid": chatid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"Message sent successfully to group {chatid}."
        except Exception as e:
            return f"Failed to send group message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_link_message(
        self,
        userid: str,
        title: str,
        text: str,
        message_url: str,
        pic_url: Optional[str] = None,
    ) -> str:
        r"""Sends a link message to a Dingtalk user.

        Args:
            userid (str): The user's userid.
            title (str): Link title.
            text (str): Link description text.
            message_url (str): URL to link to.
            pic_url (Optional[str]): Picture URL for the link.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-single-chat-message
        """
        link_data: Dict[str, Any] = {
            "title": title,
            "text": text,
            "messageUrl": message_url,
        }

        if pic_url:
            link_data["picUrl"] = pic_url

        payload = {
            "msg": {"msgtype": "link", "link": link_data},
            "userid": userid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"Link message sent successfully to user {userid}."
        except Exception as e:
            return f"Failed to send link message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_action_card_message(
        self,
        userid: str,
        title: str,
        text: str,
        single_title: str,
        single_url: str,
    ) -> str:
        r"""Sends an action card message to a Dingtalk user.

        Args:
            userid (str): The user's userid.
            title (str): Card title.
            text (str): Card content text.
            single_title (str): Action button title.
            single_url (str): Action button URL.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-single-chat-message
        """
        payload = {
            "msg": {
                "msgtype": "actionCard",
                "actionCard": {
                    "title": title,
                    "text": text,
                    "singleTitle": single_title,
                    "singleURL": single_url,
                },
            },
            "userid": userid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"Action card message sent successfully to user {userid}."
        except Exception as e:
            return f"Failed to send action card message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_user_by_mobile(self, mobile: str) -> Dict[str, Any]:
        r"""Gets user information by mobile number.

        Args:
            mobile (str): User's mobile number. Should be a valid Chinese
                mobile number format (11 digits starting with 1).

        Returns:
            Dict[str, Any]: User information or error information.
        """
        # Validate mobile number format (Chinese mobile number: 11 digits
        # starting with 1)
        mobile_pattern = r'^1[3-9]\d{9}$'
        if not re.match(mobile_pattern, mobile):
            return {
                "error": "Invalid mobile number format. Expected 11 digits "
                "starting with 1 (e.g., 13800000000)."
            }

        payload = {"mobile": mobile}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/v2/user/getbymobile",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get user by mobile: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_user_by_unionid(self, unionid: str) -> Dict[str, Any]:
        r"""Gets user information by unionid.

        Args:
            unionid (str): User's unique identifier across all DingTalk
                organizations. This is a global identifier that remains
                consistent even if the user belongs to multiple DingTalk
                organizations, unlike userid which is organization-specific.

        Returns:
            Dict[str, Any]: User information or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-a-user-by-the-union-id
        """
        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/user/getbyunionid",
                headers={"Content-Type": "application/json"},
                json={"unionid": unionid},
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get user by unionid: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_department_detail(self, dept_id: int) -> Dict[str, Any]:
        r"""Gets detailed information about a department.

        Args:
            dept_id (int): Department ID.

        Returns:
            Dict[str, Any]: Department details or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-department-details0-v2
        """
        payload = {"dept_id": dept_id}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/v2/department/get",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get department detail: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_oa_message(
        self,
        userid: str,
        message_url: str,
        head_bgcolor: str,
        head_text: str,
        body_title: str,
        body_content: str,
    ) -> str:
        r"""Sends an OA (Office Automation) message to a Dingtalk user.

        Args:
            userid (str): The user's userid.
            message_url (str): URL for the message action.
            head_bgcolor (str): Header background color (hex format).
            head_text (str): Header text.
            body_title (str): Body title.
            body_content (str): Body content.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/send-single-chat-message
        """
        oa_data: Dict[str, Any] = {
            "message_url": message_url,
            "head": {
                "bgcolor": head_bgcolor,
                "text": head_text,
            },
            "body": {
                "title": body_title,
                "form": [{"key": "Content:", "value": body_content}],
            },
        }

        payload = {
            "msg": {"msgtype": "oa", "oa": oa_data},
            "userid": userid,
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return f"OA message sent successfully to user {userid}."
        except Exception as e:
            return f"Failed to send OA message: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_group_info(self, chatid: str) -> Dict[str, Any]:
        r"""Gets information about a group chat.

        Args:
            chatid (str): Group chat ID.

        Returns:
            Dict[str, Any]: Group information or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-group-session-information
        """
        payload = {"chatid": chatid}

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/im/chat/get",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to get group info: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_update_group(
        self,
        chatid: str,
        name: Optional[str] = None,
        owner: Optional[str] = None,
        add_useridlist: Optional[List[str]] = None,
        del_useridlist: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        r"""Updates a Dingtalk group configuration.

        Args:
            chatid (str): Group chat ID.
            name (Optional[str]): New group name.
            owner (Optional[str]): New group owner userid.
            add_useridlist (Optional[List[str]]): List of user IDs to add.
                Note: Internally converted to comma-separated string as
                required by the DingTalk API.
            del_useridlist (Optional[List[str]]): List of user IDs to remove.
                Note: Internally converted to comma-separated string as
                required by the DingTalk API.

        Returns:
            Dict[str, Any]: Update result or error information.

        References:
            https://open.dingtalk.com/document/orgapp-server/modify-group-session
        """
        payload: Dict[str, Any] = {"chatid": chatid}

        if name:
            payload["name"] = name
        if owner:
            payload["owner"] = owner
        # Note: DingTalk update group API requires comma-separated string
        # format
        # This is different from send_work_notification which uses array format
        if add_useridlist:
            payload["add_useridlist"] = ",".join(add_useridlist)
        if del_useridlist:
            payload["del_useridlist"] = ",".join(del_useridlist)

        try:
            data = _make_dingtalk_request(
                "POST",
                "/topapi/im/chat/update",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return data
        except Exception as e:
            return {"error": f"Failed to update group: {e!s}"}

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_send_work_notification(
        self,
        userid_list: List[str],
        msg_content: str,
        msg_type: Literal["text", "markdown"] = "text",
    ) -> str:
        r"""Sends work notification to multiple users.

        Args:
            userid_list (List[str]): List of user IDs to send to.
                Note: This API accepts array format, unlike some other APIs
                that require comma-separated strings.
            msg_content (str): Message content.
            msg_type (Literal["text", "markdown"]): Message type.

        Returns:
            str: Success or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/asynchronous-sending-of-enterprise-session-messages
        """
        if not userid_list:
            return "Error: userid_list cannot be empty"

        if len(userid_list) > 100:
            return "Error: Cannot send to more than 100 users at once"

        msg_data: Dict[str, Any] = {"msgtype": msg_type}
        if msg_type == "text":
            msg_data["text"] = {"content": msg_content}
        elif msg_type == "markdown":
            msg_data["markdown"] = {"text": msg_content}

        payload = {
            "msg": msg_data,
            "userid_list": userid_list,  # Array format for this API
        }

        try:
            _make_dingtalk_request(
                "POST",
                "/topapi/message/corpconversation/asyncsend_v2",
                headers={"Content-Type": "application/json"},
                json=payload,
            )
            return (
                f"Work notification sent successfully to "
                f"{len(userid_list)} users."
            )
        except Exception as e:
            return f"Failed to send work notification: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_userid_by_phone(self, phone_number: str) -> str:
        r"""Gets user ID by phone number for LLM agents.

        Args:
            phone_number (str): User's phone number.

        Returns:
            str: User ID or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-user-details
        """
        try:
            user_info = self.dingtalk_get_user_by_mobile(phone_number)
            if 'result' in user_info and 'userid' in user_info['result']:
                return user_info['result']['userid']
            else:
                return f"User not found for phone number: {phone_number}"
        except Exception as e:
            return f"Failed to get user ID by phone: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_userid_by_name(self, user_name: str) -> str:
        r"""Gets user ID by user name for LLM agents.

        Args:
            user_name (str): User's display name.

        Returns:
            str: User ID or error message.

        References:
            https://open.dingtalk.com/document/orgapp-server/query-users
        """
        try:
            search_result = self.dingtalk_search_users_by_name(user_name)
            if search_result.get('result'):
                # Return the first match
                return search_result['result'][0].get(
                    'userid', f"No userid found for user: {user_name}"
                )
            else:
                return f"User not found with name: {user_name}"
        except Exception as e:
            return f"Failed to get user ID by name: {e!s}"

    @api_keys_required(
        [
            (None, "DINGTALK_APP_KEY"),
            (None, "DINGTALK_APP_SECRET"),
        ]
    )
    def dingtalk_get_department_id_by_name(self, department_name: str) -> str:
        r"""Gets department ID by department name for LLM agents.

        Args:
            department_name (str): Department name to search for.

        Returns:
            str: Department ID or error message.
        """
        try:
            dept_list = self.dingtalk_get_department_list()
            if 'result' in dept_list:
                for dept in dept_list['result']:
                    if dept.get('name') == department_name:
                        return str(dept.get('id', ''))
                return f"Department not found: {department_name}"
            else:
                return "Failed to get department list"
        except Exception as e:
            return f"Failed to get department ID by name: {e!s}"

    def dingtalk_get_chatid_by_group_name(self, group_name: str) -> str:
        r"""Gets chat ID by group name for LLM agents.

        Note: This function provides guidance as Dingtalk API doesn't directly
        support searching groups by name. Users should use the group creation
        response or group management features to obtain chat IDs.

        Args:
            group_name (str): Group name to search for.

        Returns:
            str: Guidance message for obtaining chat ID.
        """
        return (
            f"To get chat ID for group '{group_name}': "
            "1. Use dingtalk_create_group() and save the returned chatid, or "
            "2. Contact group admin to provide the chat ID, or "
            "3. Use Dingtalk admin console to find the group ID. "
            "Chat IDs are typically returned when creating groups."
        )

    def get_tools(self) -> List[FunctionTool]:
        r"""Returns toolkit functions as tools."""
        return [
            # Original message functions
            FunctionTool(self.dingtalk_send_text_message),
            FunctionTool(self.dingtalk_send_markdown_message),
            FunctionTool(self.dingtalk_send_webhook_message),
            # New message types
            FunctionTool(self.dingtalk_send_link_message),
            FunctionTool(self.dingtalk_send_action_card_message),
            FunctionTool(self.dingtalk_send_oa_message),
            FunctionTool(self.dingtalk_send_work_notification),
            # User management functions
            FunctionTool(self.dingtalk_get_user_info),
            FunctionTool(self.dingtalk_get_user_by_mobile),
            FunctionTool(self.dingtalk_get_user_by_unionid),
            FunctionTool(self.dingtalk_search_users_by_name),
            # Department management functions
            FunctionTool(self.dingtalk_get_department_list),
            FunctionTool(self.dingtalk_get_department_users),
            FunctionTool(self.dingtalk_get_department_detail),
            # Group management functions
            FunctionTool(self.dingtalk_create_group),
            FunctionTool(self.dingtalk_send_group_message),
            FunctionTool(self.dingtalk_get_group_info),
            FunctionTool(self.dingtalk_update_group),
            # Helper functions for LLM agents to get IDs
            FunctionTool(self.dingtalk_get_userid_by_phone),
            FunctionTool(self.dingtalk_get_userid_by_name),
            FunctionTool(self.dingtalk_get_department_id_by_name),
            FunctionTool(self.dingtalk_get_chatid_by_group_name),
        ]


# many other functionalities, cards etc build on top after initial works



# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========

from typing import Any, ClassVar, Dict, List, Optional, Set

from camel.toolkits import FunctionTool

from .mcp_toolkit import MCPToolkit


class NotionMCPToolkit(MCPToolkit):
    r"""NotionMCPToolkit provides an interface for interacting with Notion
    through the Model Context Protocol (MCP).

    Attributes:
        timeout (Optional[float]): Connection timeout in seconds.
            (default: :obj:`None`)

    Note:
        Currently only supports asynchronous operation mode.
    """

    # TODO: Create unified method to validate and fix the schema
    SCHEMA_KEYWORDS: ClassVar[Set[str]] = {
        "type",
        "properties",
        "items",
        "required",
        "additionalProperties",
        "description",
        "title",
        "default",
        "enum",
        "const",
        "examples",
        "$ref",
        "$defs",
        "definitions",
        "allOf",
        "oneOf",
        "anyOf",
        "not",
        "if",
        "then",
        "else",
        "format",
        "pattern",
        "minimum",
        "maximum",
        "minLength",
        "maxLength",
        "minItems",
        "maxItems",
        "uniqueItems",
    }

    def __init__(
        self,
        timeout: Optional[float] = None,
    ) -> None:
        r"""Initializes the NotionMCPToolkit.

        Args:
            timeout (Optional[float]): Connection timeout in seconds.
                (default: :obj:`None`)
        """
        config_dict = {
            "mcpServers": {
                "notionMCP": {
                    "command": "npx",
                    "args": [
                        "-y",
                        "mcp-remote",
                        "https://mcp.notion.com/mcp",
                    ],
                }
            }
        }

        # Initialize parent MCPToolkit with Notion configuration
        super().__init__(config_dict=config_dict, timeout=timeout)

    def get_tools(self) -> List[FunctionTool]:
        r"""Returns a list of tools provided by the NotionMCPToolkit.

        Returns:
            List[FunctionTool]: List of available tools.
        """
        all_tools = []
        for client in self.clients:
            try:
                original_build_schema = client._build_tool_schema

                def create_wrapper(orig_func):
                    def wrapper(mcp_tool):
                        return self._build_notion_tool_schema(
                            mcp_tool, orig_func
                        )

                    return wrapper

                client._build_tool_schema = create_wrapper(  # type: ignore[method-assign]
                    original_build_schema
                )

                client_tools = client.get_tools()
                all_tools.extend(client_tools)

                client._build_tool_schema = original_build_schema  # type: ignore[method-assign]

            except Exception as e:
                from camel.logger import get_logger

                logger = get_logger(__name__)
                logger.error(f"Failed to get tools from client: {e}")

        return all_tools

    def _build_notion_tool_schema(self, mcp_tool, original_build_schema):
        r"""Build tool schema with Notion-specific fixes."""
        schema = original_build_schema(mcp_tool)
        self._fix_notion_schema_recursively(schema)
        return schema

    def _fix_notion_schema_recursively(self, obj: Any) -> None:
        r"""Recursively fix Notion MCP schema issues."""
        if isinstance(obj, dict):
            self._fix_dict_schema(obj)
            self._process_nested_structures(obj)
        elif isinstance(obj, list):
            for item in obj:
                self._fix_notion_schema_recursively(item)

    def _fix_dict_schema(self, obj: Dict[str, Any]) -> None:
        r"""Fix dictionary schema issues."""
        if "properties" in obj and "type" not in obj:
            self._fix_missing_type_with_properties(obj)
        elif obj.get("type") == "object" and "properties" in obj:
            self._fix_object_with_properties(obj)

    def _fix_missing_type_with_properties(self, obj: Dict[str, Any]) -> None:
        r"""Fix objects with properties but missing type field."""
        properties = obj.get("properties", {})
        if properties and isinstance(properties, dict):
            obj["type"] = "object"
            obj["additionalProperties"] = False

            required_properties = self._get_required_properties(
                properties, conservative=True
            )
            if required_properties:
                obj["required"] = required_properties

    def _fix_object_with_properties(self, obj: Dict[str, Any]) -> None:
        r"""Fix objects with type="object" and properties."""
        properties = obj.get("properties", {})
        if properties and isinstance(properties, dict):
            existing_required = obj.get("required", [])

            for prop_name, prop_schema in properties.items():
                if (
                    prop_name not in existing_required
                    and prop_name not in self.SCHEMA_KEYWORDS
                    and self._is_property_required(prop_schema)
                ):
                    existing_required.append(prop_name)

            if existing_required:
                obj["required"] = existing_required

            if "additionalProperties" not in obj:
                obj["additionalProperties"] = False

    def _get_required_properties(
        self, properties: Dict[str, Any], conservative: bool = False
    ) -> List[str]:
        r"""Get list of required properties from a properties dict."""
        required = []
        for prop_name, prop_schema in properties.items():
            if (
                prop_name not in self.SCHEMA_KEYWORDS
                and isinstance(prop_schema, dict)
                and self._is_property_required(prop_schema)
            ):
                required.append(prop_name)
        return required

    def _is_property_required(self, prop_schema: Dict[str, Any]) -> bool:
        r"""Check if a property should be marked as required."""
        prop_type = prop_schema.get("type")
        return (
            prop_type is not None
            and prop_type != "null"
            and "default" not in prop_schema
            and not (isinstance(prop_type, list) and "null" in prop_type)
        )

    def _process_nested_structures(self, obj: Dict[str, Any]) -> None:
        r"""Process all nested structures in a schema object."""
        for key, value in obj.items():
            if key in ["anyOf", "oneOf", "allOf"] and isinstance(value, list):
                for item in value:
                    self._fix_notion_schema_recursively(item)
            elif key == "items" and isinstance(value, dict):
                self._fix_notion_schema_recursively(value)
            elif key == "properties" and isinstance(value, dict):
                for prop_value in value.values():
                    self._fix_notion_schema_recursively(prop_value)
            elif key == "$defs" and isinstance(value, dict):
                for def_value in value.values():
                    self._fix_notion_schema_recursively(def_value)
            elif isinstance(value, (dict, list)):
                self._fix_notion_schema_recursively(value)



# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
import json
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from camel.logger import get_logger
from camel.types.agents import ToolCallingRecord

logger = get_logger(__name__)


class WorkforceLogger:
    r"""Logs events and metrics for a Workforce instance."""

    def __init__(self, workforce_id: str):
        """Initializes the WorkforceLogger.

        Args:
            workforce_id (str): The unique identifier for the workforce.
        """
        self.workforce_id: str = workforce_id
        self.log_entries: List[Dict[str, Any]] = []
        self._task_hierarchy: Dict[str, Dict[str, Any]] = {}
        self._worker_information: Dict[str, Dict[str, Any]] = {}
        self._initial_worker_logs: List[Dict[str, Any]] = []

    def _log_event(self, event_type: str, **kwargs: Any) -> None:
        r"""Internal method to create and store a log entry.

        Args:
            event_type (str): The type of event being logged.
            **kwargs: Additional data associated with the event.
        """
        log_entry = {
            'timestamp': datetime.now(timezone.utc).isoformat(),
            'workforce_id': self.workforce_id,
            'event_type': event_type,
            **kwargs,
        }
        self.log_entries.append(log_entry)
        if event_type == 'worker_created':
            self._initial_worker_logs.append(log_entry)

    def log_task_created(
        self,
        task_id: str,
        description: str,
        parent_task_id: Optional[str] = None,
        task_type: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the creation of a new task."""
        self._log_event(
            'task_created',
            task_id=task_id,
            description=description,
            parent_task_id=parent_task_id,
            task_type=task_type,
            metadata=metadata or {},
        )
        self._task_hierarchy[task_id] = {
            'parent': parent_task_id,
            'children': [],
            'status': 'created',
            'description': description,
            'assigned_to': None,
            **(metadata or {}),
        }
        if parent_task_id and parent_task_id in self._task_hierarchy:
            self._task_hierarchy[parent_task_id]['children'].append(task_id)

    def log_task_decomposed(
        self,
        parent_task_id: str,
        subtask_ids: List[str],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the decomposition of a task into subtasks."""
        self._log_event(
            'task_decomposed',
            parent_task_id=parent_task_id,
            subtask_ids=subtask_ids,
            metadata=metadata or {},
        )
        if parent_task_id in self._task_hierarchy:
            self._task_hierarchy[parent_task_id]['status'] = "decomposed"

    def log_task_assigned(
        self,
        task_id: str,
        worker_id: str,
        queue_time_seconds: Optional[float] = None,
        dependencies: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the assignment of a task to a worker."""
        self._log_event(
            'task_assigned',
            task_id=task_id,
            worker_id=worker_id,
            queue_time_seconds=queue_time_seconds,
            dependencies=dependencies or [],
            metadata=metadata or {},
        )
        if task_id in self._task_hierarchy:
            self._task_hierarchy[task_id]['status'] = 'assigned'
            self._task_hierarchy[task_id]['assigned_to'] = worker_id
            self._task_hierarchy[task_id]['dependencies'] = dependencies or []
        if worker_id in self._worker_information:
            self._worker_information[worker_id]['current_task_id'] = task_id
            self._worker_information[worker_id]['status'] = 'busy'

    def log_task_started(
        self,
        task_id: str,
        worker_id: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs when a worker starts processing a task."""
        self._log_event(
            'task_started',
            task_id=task_id,
            worker_id=worker_id,
            metadata=metadata or {},
        )
        if task_id in self._task_hierarchy:
            self._task_hierarchy[task_id]['status'] = 'processing'

    def log_task_completed(
        self,
        task_id: str,
        worker_id: str,
        result_summary: Optional[str] = None,
        processing_time_seconds: Optional[float] = None,
        token_usage: Optional[Dict[str, int]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the successful completion of a task."""
        self._log_event(
            'task_completed',
            task_id=task_id,
            worker_id=worker_id,
            result_summary=result_summary,
            processing_time_seconds=processing_time_seconds,
            token_usage=token_usage or {},
            metadata=metadata or {},
        )
        if task_id in self._task_hierarchy:
            self._task_hierarchy[task_id]['status'] = 'completed'
            self._task_hierarchy[task_id]['assigned_to'] = None
            # Store processing time in task hierarchy for display in tree
            if processing_time_seconds is not None:
                self._task_hierarchy[task_id]['completion_time_seconds'] = (
                    processing_time_seconds
                )
            # Store token usage in task hierarchy for display in tree
            if token_usage is not None:
                self._task_hierarchy[task_id]['token_usage'] = token_usage
        if worker_id in self._worker_information:
            self._worker_information[worker_id]['current_task_id'] = None
            self._worker_information[worker_id]['status'] = 'idle'
            self._worker_information[worker_id]['tasks_completed'] = (
                self._worker_information[worker_id].get('tasks_completed', 0)
                + 1
            )

    def log_task_failed(
        self,
        task_id: str,
        error_message: str,
        worker_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the failure of a task."""
        self._log_event(
            'task_failed',
            task_id=task_id,
            worker_id=worker_id,
            error_message=error_message,
            metadata=metadata or {},
        )
        if task_id in self._task_hierarchy:
            self._task_hierarchy[task_id]['status'] = 'failed'
            self._task_hierarchy[task_id]['error'] = error_message
            self._task_hierarchy[task_id]['assigned_to'] = None
        if worker_id and worker_id in self._worker_information:
            self._worker_information[worker_id]['current_task_id'] = None
            self._worker_information[worker_id]['status'] = 'idle'
            self._worker_information[worker_id]['tasks_failed'] = (
                self._worker_information[worker_id].get('tasks_failed', 0) + 1
            )

    def log_worker_created(
        self,
        worker_id: str,
        worker_type: str,
        role: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the creation of a new worker."""
        self._log_event(
            'worker_created',
            worker_id=worker_id,
            worker_type=worker_type,
            role=role,
            metadata=metadata or {},
        )
        self._worker_information[worker_id] = {
            'type': worker_type,
            'role': role,
            'status': 'idle',
            'current_task_id': None,
            'tasks_completed': 0,
            'tasks_failed': 0,
            **(metadata or {}),
        }

    def log_worker_deleted(
        self,
        worker_id: str,
        reason: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the deletion of a worker."""
        self._log_event(
            'worker_deleted',
            worker_id=worker_id,
            reason=reason,
            metadata=metadata or {},
        )
        if worker_id in self._worker_information:
            self._worker_information[worker_id]['status'] = 'deleted'
            # Or del self._worker_information[worker_id]

    def reset_task_data(self) -> None:
        r"""Resets logs and data related to tasks, preserving worker
        information.
        """
        # Restore log entries from the initial worker logs
        self.log_entries = list(self._initial_worker_logs)  # Make a copy

        self._task_hierarchy.clear()
        for worker_id in self._worker_information:
            if (
                self._worker_information[worker_id].get('status') != 'deleted'
            ):  # Don't revive deleted workers
                self._worker_information[worker_id]['current_task_id'] = None
                self._worker_information[worker_id]['status'] = 'idle'
        logger.info(
            f"WorkforceLogger: Task data reset for workforce "
            f"{self.workforce_id}"
        )

    def log_queue_status(
        self,
        queue_name: str,
        length: int,
        pending_task_ids: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Logs the status of a task queue."""
        self._log_event(
            'queue_status',
            queue_name=queue_name,
            length=length,
            pending_task_ids=pending_task_ids or [],
            metadata=metadata or {},
        )

    def dump_to_json(self, file_path: str) -> None:
        r"""Dumps all log entries to a JSON file.

        Args:
            file_path (str): The path to the JSON file.
        """

        def json_serializer_default(o: Any) -> Any:
            if isinstance(o, ToolCallingRecord):
                return o.as_dict()
            # Let the default encoder raise the TypeError for other types
            raise TypeError(
                f"Object of type {o.__class__.__name__} is not "
                f"JSON serializable"
            )

        try:
            with open(file_path, 'w') as f:
                json.dump(
                    self.log_entries,
                    f,
                    indent=4,
                    default=json_serializer_default,
                )
        except IOError as e:
            # Consider using camel.logger for this kind of internal error
            logger.error(f"Error dumping logs to JSON: {e}")

    def _get_all_tasks_in_hierarchy(
        self, task_id: str
    ) -> Dict[str, Dict[str, Any]]:
        r"""Recursively collect all tasks in the hierarchy starting from
        task_id.
        """
        result: Dict[str, Dict[str, Any]] = {}
        if task_id not in self._task_hierarchy:
            return result

        # Add the current task
        result[task_id] = self._task_hierarchy[task_id]

        # Add all children recursively
        children = self._task_hierarchy[task_id].get('children', [])
        for child_id in children:
            result.update(self._get_all_tasks_in_hierarchy(child_id))

        return result

    def _get_task_tree_string(
        self, task_id: str, prefix: str = "", is_last: bool = True
    ) -> str:
        r"""Generate a string representation of the task tree."""
        if task_id not in self._task_hierarchy:
            return ""

        task_info = self._task_hierarchy[task_id]
        description = task_info.get('description', '')
        status = task_info.get('status', 'unknown')
        assignee = task_info.get('assigned_to')
        assignee_str = f" [assigned to: {assignee}]" if assignee else ""
        dependencies = task_info.get('dependencies', [])
        dependencies_list = [
            dep for dep in dependencies if dep in self._task_hierarchy
        ]
        dependencies_str = (
            f" (dependencies: {', '.join(dependencies_list)})"
            if dependencies_list
            else ""
        )
        error_str = (
            f" [ERROR: {task_info.get('error', '')}]"
            if status == 'failed'
            else ""
        )

        # Add completion time and token usage for completed tasks
        completion_time_str = ""
        token_usage_str = ""

        if status == 'completed':
            # For the root task (typically task_id = '0'), calculate total
            # tokens and time
            if task_id == '0':
                # Calculate total tokens from all child tasks
                total_tokens = 0
                total_time = 0.0

                # Recursively get all tasks in the hierarchy
                all_tasks = self._get_all_tasks_in_hierarchy(task_id)

                # Sum up tokens and time from all tasks
                for child_id, child_info in all_tasks.items():
                    if (
                        child_id != task_id
                    ):  # Skip the root task itself to avoid double counting
                        # Add tokens
                        if (
                            'token_usage' in child_info
                            and child_info['token_usage'] is not None
                        ):
                            child_tokens = child_info['token_usage']
                            if (
                                isinstance(child_tokens, dict)
                                and 'total_tokens' in child_tokens
                            ):
                                total_tokens += child_tokens['total_tokens']
                            elif isinstance(child_tokens, int):
                                total_tokens += child_tokens

                        # Add completion time
                        if (
                            'completion_time_seconds' in child_info
                            and child_info['completion_time_seconds']
                            is not None
                        ):
                            total_time += child_info['completion_time_seconds']

                # Format the strings for the root task
                completion_time_str = (
                    f" (completed in {total_time:.2f} seconds total)"
                )
                token_usage_str = f" [total tokens: {total_tokens}]"
            else:
                # Regular task (not root) - show its own completion time and
                # tokens
                if (
                    'completion_time_seconds' in task_info
                    and task_info['completion_time_seconds'] is not None
                ):
                    completion_time = task_info['completion_time_seconds']
                    completion_time_str = (
                        f" (completed in {completion_time:.2f} seconds)"
                    )
                else:
                    # Add a default message when completion time is not
                    # available
                    completion_time_str = " (completed)"

                # Add token usage if available
                if (
                    'token_usage' in task_info
                    and task_info['token_usage'] is not None
                ):
                    token_usage = task_info['token_usage']
                    if (
                        isinstance(token_usage, dict)
                        and 'total_tokens' in token_usage
                    ):
                        token_usage_str = (
                            f" [tokens: {token_usage['total_tokens']}]"
                        )
                    elif isinstance(token_usage, int):
                        token_usage_str = f" [tokens: {token_usage}]"

        tree_str = f"{prefix}{'`-- ' if is_last else '|-- '}[{task_id}] {description} [{status}]{completion_time_str}{token_usage_str}{assignee_str}{dependencies_str}{error_str}\n"  # noqa: E501

        children = task_info.get('children', [])
        for i, child_id in enumerate(children):
            new_prefix = prefix + ("    " if is_last else "|   ")
            tree_str += self._get_task_tree_string(
                child_id, new_prefix, i == len(children) - 1
            )
        return tree_str

    def get_ascii_tree_representation(self) -> str:
        r"""Generates an ASCII tree representation of the current task
        hierarchy and worker status.
        """
        output_str = "=== Task Hierarchy ===\n"
        root_tasks = [
            task_id
            for task_id, info in self._task_hierarchy.items()
            if info.get('parent') is None
        ]
        if not root_tasks:
            output_str += "No tasks recorded.\n"
        else:
            for i, task_id in enumerate(root_tasks):
                output_str += self._get_task_tree_string(
                    task_id, "", i == len(root_tasks) - 1
                )

        output_str += "\n=== Worker Information ===\n"
        if not self._worker_information:
            output_str += "No workers recorded.\n"
        else:
            for worker_id, info in self._worker_information.items():
                role = info.get('role', 'N/A')
                completed = info.get('tasks_completed', 0)
                failed = info.get('tasks_failed', 0)
                output_str += (
                    f"- Worker ID: {worker_id} (Role: {role})\n"
                    f"    Tasks Completed: {completed}, Tasks "
                    f"Failed: {failed}\n"
                )
        return output_str

    def get_kpis(self) -> Dict[str, Any]:
        r"""Calculates and returns key performance indicators from the logs."""
        kpis: Dict[str, Any] = {
            'total_tasks_created': 0,
            'total_tasks_completed': 0,
            'total_tasks_failed': 0,
            'worker_utilization': {},
            'current_pending_tasks': 0,
            'total_workforce_running_time_seconds': 0.0,
        }

        task_start_times: Dict[str, float] = {}
        task_creation_timestamps: Dict[str, datetime] = {}
        task_assignment_timestamps: Dict[str, datetime] = {}
        first_timestamp: Optional[datetime] = None
        last_timestamp: Optional[datetime] = None

        tasks_handled_by_worker: Dict[str, int] = {}

        # Helper function to check if a task is the main task (has no parent)
        def is_main_task(task_id: str) -> bool:
            return (
                task_id in self._task_hierarchy
                and self._task_hierarchy[task_id].get('parent') is None
            )

        for entry in self.log_entries:
            event_type = entry['event_type']
            timestamp = datetime.fromisoformat(entry['timestamp'])
            task_id = entry.get('task_id', '')

            if first_timestamp is None or timestamp < first_timestamp:
                first_timestamp = timestamp
            if last_timestamp is None or timestamp > last_timestamp:
                last_timestamp = timestamp

            if event_type == 'task_created':
                # Exclude main task from total count
                if not is_main_task(task_id):
                    kpis['total_tasks_created'] += 1
                task_creation_timestamps[task_id] = timestamp
            elif event_type == 'task_assigned':
                task_assignment_timestamps[task_id] = timestamp
                # Queue time tracking has been removed

            elif event_type == 'task_started':
                # Store start time for processing time calculation
                task_start_times[task_id] = timestamp.timestamp()

            elif event_type == 'task_completed':
                # Exclude main task from total count
                if not is_main_task(task_id):
                    kpis['total_tasks_completed'] += 1
                    # Count tasks handled by worker (only for non-main tasks)
                    if 'worker_id' in entry and entry['worker_id'] is not None:
                        worker_id = entry['worker_id']
                        tasks_handled_by_worker[worker_id] = (
                            tasks_handled_by_worker.get(worker_id, 0) + 1
                        )

                if task_id in task_assignment_timestamps:
                    completion_time = (
                        timestamp - task_assignment_timestamps[task_id]
                    ).total_seconds()
                    # Store completion time in task hierarchy instead of KPIs
                    # array
                    if task_id in self._task_hierarchy:
                        self._task_hierarchy[task_id][
                            'completion_time_seconds'
                        ] = completion_time

            elif event_type == 'task_failed':
                # Exclude main task from total count
                if not is_main_task(task_id):
                    kpis['total_tasks_failed'] += 1
                    # Count tasks handled by worker (only for non-main tasks)
                    if 'worker_id' in entry and entry['worker_id'] is not None:
                        worker_id = entry['worker_id']
                        tasks_handled_by_worker[worker_id] = (
                            tasks_handled_by_worker.get(worker_id, 0) + 1
                        )
            elif event_type == 'queue_status':
                pass  # Placeholder for now

        # Calculate total workforce running time
        if first_timestamp and last_timestamp and self.log_entries:
            kpis['total_workforce_running_time_seconds'] = (
                last_timestamp - first_timestamp
            ).total_seconds()
        # Calculate worker utilization based on proportion of tasks handled
        total_tasks_processed_for_utilization = (
            kpis['total_tasks_completed'] + kpis['total_tasks_failed']
        )
        if total_tasks_processed_for_utilization > 0:
            for (
                worker_id_key,
                num_tasks_handled,
            ) in tasks_handled_by_worker.items():
                percentage = (
                    num_tasks_handled / total_tasks_processed_for_utilization
                ) * 100
                kpis['worker_utilization'][worker_id_key] = (
                    f"{percentage:.2f}%"
                )
        else:
            for worker_id_key in (
                tasks_handled_by_worker
            ):  # Ensure all workers who handled tasks are listed, even if 0%
                kpis['worker_utilization'][worker_id_key] = "0.00%"
            # If no tasks were processed, but workers exist (e.g. from
            # _initial_worker_logs), list them with 0%
            for worker_id_key in self._worker_information:
                if worker_id_key not in kpis['worker_utilization']:
                    kpis['worker_utilization'][worker_id_key] = "0.00%"

        # Task throughput (completed tasks per minute, for example)
        if self.log_entries:
            first_log_time = datetime.fromisoformat(
                self.log_entries[0]['timestamp']
            )
            last_log_time = datetime.fromisoformat(
                self.log_entries[-1]['timestamp']
            )
            duration_seconds = (last_log_time - first_log_time).total_seconds()
            if duration_seconds > 0:
                kpis['task_throughput_per_second'] = (
                    kpis['total_tasks_completed'] / duration_seconds
                )
                kpis['task_throughput_per_minute'] = (
                    kpis['task_throughput_per_second'] * 60
                )

        kpis['total_workers_created'] = len(self._worker_information)

        # Current pending tasks (simplified)
        kpis['current_pending_tasks'] = kpis['total_tasks_created'] - (
            kpis['total_tasks_completed'] + kpis['total_tasks_failed']
        )

        return kpis
